package handler

import (
	"database/sql"
	"encoding/base64"
	"encoding/json"
	"github.com/mholt/binding"
	"github.com/violendy/SalesPortal/internal/config"
	"github.com/violendy/SalesPortal/internal/models"
	"github.com/violendy/SalesPortal/internal/repository"
	"github.com/violendy/SalesPortal/internal/repository/apirepo"
	"github.com/violendy/SalesPortal/internal/repository/dbrepo"
	"log"
	"net/http"
	"os"
	"strconv"
	"time"
)

//TODO
//	- Implement sales crud
//  - Implement user crud
//  - Implement local sync

type TLocalSalesAuthCache struct {
	AuthCache     map[string]models.SalesData
	LastEntryDate time.Time
}

type TLocalUserAuthCache struct {
	AuthCache     map[string]models.UserData
	LastEntryDate time.Time
}

var Repo *Repository
var LocalSalesAuthCache TLocalSalesAuthCache
var LocalUserAuthCache TLocalUserAuthCache

type Repository struct {
	App      *config.AppConfig
	PortalDB repository.PortalDatabaseRepo
	NKDB     repository.NKDatabaseRepo
	WA       repository.WAGatewayRepo
}

//	func (rp *Repository) EnableCors(w *http.ResponseWriter) {
//		(*w).Header().Set("Access-Control-Allow-Origin", "*")
//
//		(*w).Header().Set("Access-Control-Allow-Methods", "GET, POST, PUT, DELETE, OPTIONS")
//
//		(*w).Header().Set("Access-Control-Allow-Headers", ",Authorization, Content-Type")
//	}
func (rp *Repository) ServerResponse(w http.ResponseWriter, msg string, datalength int, data interface{}) {
	w.Header().Set("Content-Type", "application/json")
	w.WriteHeader(http.StatusOK)
	var ServerResponse = models.GeneralServerResponse{
		Message:            msg,
		ResponseData:       data,
		ResponseDataLength: &datalength,
		HTTPStatusCode:     http.StatusOK,
		Error:              false,
	}
	err := json.NewEncoder(w).Encode(ServerResponse)
	if err != nil {
		log.Print(err)
		return
	}

}
func (rp *Repository) ServerErrorResponse(w http.ResponseWriter, ErrorMessage string, DebugMessage string, DebugErrMessage string, DebugData interface{}, Code int) {
	w.Header().Set("Content-Type", "application/json")
	w.WriteHeader(http.StatusOK)
	var ErrorResponse = models.GeneralServerResponse{
		ErrorMessage:   ErrorMessage,
		Error:          true,
		HTTPStatusCode: Code,
	}
	if !rp.App.InProduction {
		ErrorResponse.DebugMessage = DebugMessage
		ErrorResponse.DebugErrMessage = DebugErrMessage
		ErrorResponse.DebugData = DebugData
	}

	err := json.NewEncoder(w).Encode(ErrorResponse)
	if err != nil {
		log.Print(err)
		return
	}
}
func NewRepo(a *config.AppConfig, db *sql.DB, dblist map[string]*sql.DB) *Repository {
	LocalSalesAuthCache.LastEntryDate = time.Now()
	LocalSalesAuthCache.AuthCache = make(map[string]models.SalesData)
	//log.Print(len(LocalSalesAuthCache.AuthCache))

	return &Repository{
		App:      a,
		PortalDB: dbrepo.NewPortalRepo(db, a),
		NKDB:     dbrepo.NewNKRepo(dblist, a),
		WA:       apirepo.NewWARepo(a),
	}
}

func NewHandlers(r *Repository) {
	Repo = r
}

type SuccessResponse struct {
	Msg        string
	DataLength int
	Data       interface{}
}

type ErrorResponse struct {
	ErrorMessage    string
	DebugMessage    string
	DebugErrMessage string
	DebugData       interface{}
	Code            int
}

type ServerResponses struct {
	Success       SuccessResponse
	TokenMissing  ErrorResponse
	TokenInvalid  ErrorResponse
	BindingError  ErrorResponse
	FunctionError ErrorResponse
}
type RequestUserFunctionWithBinding func(User models.UserData, UserStruct interface{}, params interface{}) (error, int, interface{})
type RequestUserFunction func(User models.UserData, params interface{}) (error, int, interface{})

type RequestSalesFunctionWithBinding func(Sales models.SalesData, UserStruct interface{}, params interface{}) (error, int, interface{})
type RequestSalesFunction func(Sales models.SalesData, params interface{}) (error, int, interface{})

type RESTMethod string

const (
	GET    RESTMethod = "GET"
	POST              = "POST"
	PUT               = "PUT"
	DELETE            = "DELETE"
)

func GenerateServerResponse(Entity string, Method RESTMethod) ServerResponses {
	return ServerResponses{
		Success: SuccessResponse{
			Msg:        string(Method + " request succeed."),
			DataLength: 0,
			Data:       nil,
		},
		TokenMissing: ErrorResponse{
			ErrorMessage: string(Method + " request error."),
			DebugMessage: string("error "+Method) + " " + Entity + " : no authentication header.",
			DebugData:    nil,
			Code:         http.StatusBadRequest,
		},
		TokenInvalid: ErrorResponse{
			ErrorMessage: string(Method + " request error."),
			DebugMessage: string("error "+Method) + " " + Entity + " : invalid/expired authentication header",
			DebugData:    nil,
			Code:         http.StatusBadRequest,
		},
		BindingError: ErrorResponse{
			ErrorMessage: string(Method + " request error."),
			DebugMessage: string("error "+Method) + " " + Entity + " : error binding to user struct type",
			DebugData:    nil,
			Code:         http.StatusBadRequest,
		},
		FunctionError: ErrorResponse{
			ErrorMessage: string(Method + " request error."),
			DebugMessage: "",
			DebugData:    nil,
			Code:         http.StatusInternalServerError,
		},
	}
}
func (rp *Repository) saveBase64File(w http.ResponseWriter, r *http.Request) {
	var payload struct {
		FileContent string `json:"file_content"`
	}

	err := json.NewDecoder(r.Body).Decode(&payload)
	if err != nil {
		http.Error(w, err.Error(), http.StatusBadRequest)
		return
	}

	// decode the base64-encoded file contents
	fileContents, err := base64.StdEncoding.DecodeString(payload.FileContent)
	if err != nil {
		http.Error(w, err.Error(), http.StatusBadRequest)
		return
	}

	// create destination file making sure the path is writeable.
	dst, err := os.Create("/tmp/uploadFile")
	if err != nil {
		http.Error(w, err.Error(), http.StatusInternalServerError)
		return
	}
	defer dst.Close()

	// write the binary data to the destination file
	if _, err := dst.Write(fileContents); err != nil {
		http.Error(w, err.Error(), http.StatusInternalServerError)
		return
	}

	w.Write([]byte("File uploaded successfully."))
}
func (rp *Repository) ProcessSalesRequest(w http.ResponseWriter, r *http.Request, f RequestSalesFunction, params interface{}, sr ServerResponses) {
	sAuth := r.Header.Get("Authorization")
	if sAuth != "" {
		Sales, bAuthenticated := rp.AuthenticateSales(sAuth)
		if bAuthenticated {
			err, recCount, data := f(Sales, params)
			if err != nil {
				rp.ServerErrorResponse(w, sr.FunctionError.ErrorMessage, sr.FunctionError.DebugMessage, err.Error(), nil, sr.FunctionError.Code)
				return
			} else {
				sr.Success.DataLength = recCount
				rp.ServerResponse(w, sr.Success.Msg, sr.Success.DataLength, data)
				return
			}
		} else {
			rp.ServerErrorResponse(w, sr.TokenInvalid.ErrorMessage, sr.TokenInvalid.DebugMessage, strconv.FormatBool(bAuthenticated), sr.TokenInvalid.DebugData, sr.TokenInvalid.Code)
			return
		}
	} else {
		rp.ServerErrorResponse(w, sr.TokenMissing.ErrorMessage, sr.TokenMissing.DebugMessage, "", sr.TokenMissing.DebugData, sr.TokenMissing.Code)
		return
	}
}
func (rp *Repository) ProcessSalesRequestWithBinding(w http.ResponseWriter, r *http.Request, f RequestSalesFunctionWithBinding, us interface{}, params interface{}, sr ServerResponses) {
	sAuth := r.Header.Get("Authorization")
	if sAuth != "" {
		Sales, bAuthenticated := rp.AuthenticateSales(sAuth)
		if bAuthenticated {
			var NewBinding binding.FieldMapper

			if fieldMapper, ok := us.(binding.FieldMapper); ok {
				NewBinding = fieldMapper
				if errs := binding.Bind(r, NewBinding); errs != nil {
					rp.ServerErrorResponse(w, sr.BindingError.ErrorMessage, sr.BindingError.DebugMessage, errs.Error(), NewBinding, sr.BindingError.Code)
					return
				} else {
					err, recCount, data := f(Sales, NewBinding, params)
					if err != nil {
						rp.ServerErrorResponse(w, sr.FunctionError.ErrorMessage, sr.FunctionError.DebugMessage, err.Error(), nil, sr.FunctionError.Code)
						return
					} else {
						rp.ServerResponse(w, sr.Success.Msg, recCount, data)
						return
					}
				}
			} else {
				rp.ServerErrorResponse(w, sr.BindingError.ErrorMessage, sr.BindingError.DebugMessage, "", NewBinding, sr.BindingError.Code)
				return
			}
		} else {
			rp.ServerErrorResponse(w, sr.TokenInvalid.ErrorMessage, sr.TokenInvalid.DebugMessage, strconv.FormatBool(bAuthenticated), sr.TokenInvalid.DebugData, sr.TokenInvalid.Code)
			return
		}
	} else {
		rp.ServerErrorResponse(w, sr.TokenMissing.ErrorMessage, sr.TokenMissing.DebugMessage, "", sr.TokenMissing.DebugData, sr.TokenMissing.Code)
		return
	}
}
func (rp *Repository) ProcessPortalRequestWithBinding(w http.ResponseWriter, r *http.Request, f RequestUserFunctionWithBinding, us interface{}, params interface{}, sr ServerResponses) {
	sAuth := r.Header.Get("Authorization")
	if sAuth != "" {
		User, bAuthenticated := rp.AuthenticateUser(sAuth)
		if bAuthenticated {
			var NewBinding binding.FieldMapper

			if fieldMapper, ok := us.(binding.FieldMapper); ok {
				NewBinding = fieldMapper
				if errs := binding.Bind(r, NewBinding); errs != nil {
					rp.ServerErrorResponse(w, sr.BindingError.ErrorMessage, sr.BindingError.DebugMessage, errs.Error(), NewBinding, sr.BindingError.Code)
					return
				} else {
					err, recCount, data := f(User, NewBinding, params)

					if err != nil {
						rp.ServerErrorResponse(w, sr.FunctionError.ErrorMessage, sr.FunctionError.DebugMessage, err.Error(), nil, sr.FunctionError.Code)
						return
					} else {
						rp.ServerResponse(w, sr.Success.Msg, recCount, data)
						return
					}
				}
			} else {
				rp.ServerErrorResponse(w, sr.BindingError.ErrorMessage, sr.BindingError.DebugMessage, "", NewBinding, sr.BindingError.Code)
				return
			}
		} else {
			rp.ServerErrorResponse(w, sr.TokenInvalid.ErrorMessage, sr.TokenInvalid.DebugMessage, strconv.FormatBool(bAuthenticated), sr.TokenInvalid.DebugData, sr.TokenInvalid.Code)
			return
		}
	} else {
		rp.ServerErrorResponse(w, sr.TokenMissing.ErrorMessage, sr.TokenMissing.DebugMessage, "", sr.TokenMissing.DebugData, sr.TokenMissing.Code)
		return
	}
}

func (rp *Repository) ProcessPortalRequest(w http.ResponseWriter, r *http.Request, f RequestUserFunction, params interface{}, sr ServerResponses) {
	sAuth := r.Header.Get("Authorization")
	if sAuth != "" {
		User, bAuthenticated := rp.AuthenticateUser(sAuth)
		if bAuthenticated {
			err, recCount, data := f(User, params)
			if err != nil {
				rp.ServerErrorResponse(w, sr.FunctionError.ErrorMessage, sr.FunctionError.DebugMessage, err.Error(), nil, sr.FunctionError.Code)
				return
			} else {
				sr.Success.DataLength = recCount
				rp.ServerResponse(w, sr.Success.Msg, sr.Success.DataLength, data)
				return
			}
		} else {
			rp.ServerErrorResponse(w, sr.TokenInvalid.ErrorMessage, sr.TokenInvalid.DebugMessage, strconv.FormatBool(bAuthenticated), sr.TokenInvalid.DebugData, sr.TokenInvalid.Code)
			return
		}
	} else {
		rp.ServerErrorResponse(w, sr.TokenMissing.ErrorMessage, sr.TokenMissing.DebugMessage, "", sr.TokenMissing.DebugData, sr.TokenMissing.Code)
		return
	}
}
func (rp *Repository) AuthenticateUser(sAuth string) (models.UserData, bool) {
	if time.Now().Format("2006/01/02") != LocalUserAuthCache.LastEntryDate.Format("2006/01/02") {
		LocalUserAuthCache.AuthCache = make(map[string]models.UserData)
	}
	if sAuth == "" {
		return models.UserData{}, false
	} else {
		log.Print("UserAuthCacheSize : ", len(LocalSalesAuthCache.AuthCache))
		CachedAuth, ok := LocalUserAuthCache.AuthCache[sAuth]
		if ok {
			return CachedAuth, true
		} else {
			if rp.PortalDB.UserCheckToken(sAuth) {
				User, err := rp.PortalDB.UserGetByToken(sAuth)
				if err != nil {
					return models.UserData{}, false
				}
				LocalUserAuthCache.AuthCache[sAuth] = User
				return User, true
			} else {
				return models.UserData{}, false
			}
		}
	}
}
func (rp *Repository) AuthenticateSales(sAuth string) (models.SalesData, bool) {
	if time.Now().Format("2006/01/02") != LocalSalesAuthCache.LastEntryDate.Format("2006/01/02") {
		LocalSalesAuthCache.AuthCache = make(map[string]models.SalesData)
	}
	if sAuth == "" {
		return models.SalesData{}, false
	} else {
		log.Print("AuthCacheSize : ", len(LocalSalesAuthCache.AuthCache))
		CachedAuth, ok := LocalSalesAuthCache.AuthCache[sAuth]
		if ok {
			return CachedAuth, true
		} else {
			if rp.PortalDB.SalesCheckToken(sAuth) {
				Sales, err := rp.PortalDB.SalesGetByToken(sAuth)
				if err != nil {
					return models.SalesData{}, false
				}
				LocalSalesAuthCache.AuthCache[sAuth] = Sales
				return Sales, true
			} else {
				return models.SalesData{}, false
			}
		}
	}
}
