package main

import (
	"database/sql"
	"fmt"
	"io"

	"context"
	//"encoding/gob"
	"github.com/alexedwards/scs/v2"
	"github.com/foomo/simplecert"
	"github.com/foomo/tlsconfig"
	_ "github.com/go-sql-driver/mysql"
	"github.com/violendy/SalesPortal/internal/config"
	"github.com/violendy/SalesPortal/internal/driver"
	"github.com/violendy/SalesPortal/internal/handler"
	"github.com/violendy/SalesPortal/internal/render"
	"log"
	"net/http"
	"os"
	"time"
	_ "time/tzdata"
)

const portNumber = ":80"

var app config.AppConfig
var session *scs.SessionManager

func main() {
	f, errs := os.OpenFile("log_"+time.Now().Format("060102")+".txt", os.O_WRONLY|os.O_CREATE|os.O_APPEND, 0644)
	if errs != nil {
		log.Fatal(errs)
	}
	defer f.Close()

	// Set the log output to write to the file and the screen
	log.SetOutput(io.MultiWriter(os.Stdout, f))

	var db, _ = run()

	var (
		// the structure that handles reloading the certificate
		certReloader *simplecert.CertReloader
		err          error
		numRenews    int
		ctx, cancel  = context.WithCancel(context.Background())

		// init strict tlsConfig (this will enforce the use of modern TLS configurations)
		// you could use a less strict configuration if you have a customer facing web application that has visitors with old browsers
		tlsConf = tlsconfig.NewServerTLSConfig(tlsconfig.TLSModeServerStrict)

		// a simple constructor for a http.Server with our Handler
		makeServer = func() *http.Server {
			return &http.Server{
				Addr:      ":443",
				Handler:   routes(&app),
				TLSConfig: tlsConf,
			}
		}

		// init server
		srv = makeServer()

		// init simplecert configuration
		cfg = simplecert.Default
	)
	if app.InProduction == false {

		makeServer = func() *http.Server {
			return &http.Server{
				Addr:    ":80",
				Handler: routes(&app),
			}
		}
		srv = makeServer()
	}
	//cfg.DNSProvider = "cloudflare"

	cfg.Domains = []string{"metalindo.linksoftindo.com", "portalsales.linksoftindo.com"}
	cfg.CacheDir = "letsencrypt"
	cfg.SSLEmail = "violendy@gmail.com"

	// disable HTTP challenges - we will only use the TLS challenge for this example.
	cfg.HTTPAddress = ""

	// this function will be called just before certificate renewal starts and is used to gracefully stop the service
	// (we need to temporarily free port 443 in order to complete the TLS challenge)
	cfg.WillRenewCertificate = func() {
		// stop server
		cancel()
	}

	// this function will be called after the certificate has been renewed, and is used to restart your service.
	cfg.DidRenewCertificate = func() {

		numRenews++

		// restart server: both context and server instance need to be recreated!
		ctx, cancel = context.WithCancel(context.Background())
		srv = makeServer()

		// force reload the updated cert from disk
		certReloader.ReloadNow()

		// here we go again
		go serve(ctx, srv)
	}

	// init simplecert configuration
	// this will block initially until the certificate has been obtained for the first time.
	// on subsequent runs, simplecert will load the certificate from the cache directory on disk.

	if app.InProduction {
		certReloader, err = simplecert.Init(cfg, func() {
			os.Exit(0)
		})
		if err != nil {
			log.Fatal("simplecert init failed: ", err)
		}

		// redirect HTTP to HTTPS
		log.Println("starting HTTP Listener on Port 80")
		go http.ListenAndServe(":80", http.HandlerFunc(simplecert.Redirect))
		// enable hot reload
		tlsConf.GetCertificate = certReloader.GetCertificateFunc()

		// start serving
		log.Println("will serve at: https://" + cfg.Domains[0])
	}

	serve(ctx, srv)

	fmt.Println("waiting forever")
	<-make(chan bool)
	defer func(SQL *sql.DB) {
		err = SQL.Close()
		if err != nil {
			os.Exit(1)
		}
	}(db.SQL)

}

func run() (*driver.DB, error) {
	//gob.Register(models.UserType{})

	app.InProduction = true
	if !app.InProduction {
		app.ContextTime = time.Hour
	} else {
		app.ContextTime = time.Second
	}
	session = scs.New()
	session.Lifetime = 24 * time.Hour
	session.Cookie.Persist = true
	session.Cookie.SameSite = http.SameSiteLaxMode
	session.Cookie.Secure = app.InProduction
	app.Session = session

	log.Println("Connecting to database...")
	db, err := driver.ConnectSQL("mysql", "root:@tcp(localhost:3306)/dbsalesportal?parseTime=true")
	if err != nil {
		return nil, err
	}

	//dbb2m, err := driver.ConnectSQL("mysql", "client:linksoft@tcp(besibaja-082022.ddns.net:3306)/dbfastretail?parseTime=true&allowOldPasswords=1")
	dbb2m, err := driver.ConnectSQL("mysql", "root:@tcp(localhost:3306)/dbb2m?parseTime=true&allowOldPasswords=1")
	if err != nil {
		return nil, err
	}

	dbekm, err := driver.ConnectSQL("mysql", "root:@tcp(localhost:3306)/dbsalesportal?parseTime=true&allowOldPasswords=1")
	if err != nil {
		return nil, err
	}

	dbakm, err := driver.ConnectSQL("mysql", "root:@tcp(localhost:3306)/dbsalesportal?parseTime=true")
	if err != nil {
		return nil, err
	}

	dbakmm, err := driver.ConnectSQL("mysql", "root:@tcp(localhost:3306)/dbsalesportal?parseTime=true")
	if err != nil {
		return nil, err
	}

	dbmkm, err := driver.ConnectSQL("mysql", "root:@tcp(localhost:3306)/dbsalesportal?parseTime=true")
	if err != nil {
		return nil, err
	}

	dbmkmm, err := driver.ConnectSQL("mysql", "root:@tcp(localhost:3306)/dbsalesportal?parseTime=true")
	if err != nil {
		return nil, err
	}

	log.Println("Connected to database!")

	tc, err := render.CreateTemplateCache()
	if err != nil {
		//log.Fatal("Main App : Can't create Template Cache")
		return nil, err
	}

	app.TemplateCache = tc
	app.UseCache = false
	app.ContextValue = 50
	render.NewTemplates(&app)
	var DBList = make(map[string]*sql.DB)

	DBList["100096842527670273"] = dbb2m.SQL
	DBList["mkm"] = dbmkm.SQL
	DBList["akm"] = dbakm.SQL
	DBList["ekm"] = dbekm.SQL
	DBList["akmm"] = dbakmm.SQL
	DBList["mkmm"] = dbmkmm.SQL

	Repo := handler.NewRepo(&app, db.SQL, DBList)
	handler.NewHandlers(Repo)
	return db, err
}
func serve(ctx context.Context, srv *http.Server) {
	go func() {
		if app.InProduction {
			log.Print("Listened on port " + srv.Addr)
			if err := srv.ListenAndServeTLS("", ""); err != nil && err != http.ErrServerClosed {
				log.Fatalf("listen: %+s\n", err)
			}
		} else {
			log.Print("Listened on port " + srv.Addr)
			err := http.ListenAndServe(":80", routes(&app))

			if err != nil {
				log.Fatal(err)
			}

		}
	}()

	log.Printf("server started")
	<-ctx.Done()
	log.Printf("server stopped")

	ctxShutDown, cancel := context.WithTimeout(context.Background(), 5*time.Second)
	defer func() {
		cancel()
	}()

	err := srv.Shutdown(ctxShutDown)
	if err == http.ErrServerClosed {
		log.Printf("server exited properly")
	} else if err != nil {
		log.Printf("server encountered an error on exit: %+s\n", err)
	}
}
